import { scalar } from "./scalar";
import { _type } from "./symbols";
import { mergeOptions } from "./validate";
import { vality } from "./vality";
vality.string = scalar("string", (val, options, context) => {
    if (typeof val === "string")
        return val;
    const { strict } = mergeOptions(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "number")
        return undefined;
    return val.toString();
}, {
    minLength: (val, o) => val.length >= o,
    maxLength: (val, o) => val.length <= o,
    match: (val, o) => o.test(val),
});
vality.number = scalar("number", (val, options, context) => {
    if (typeof val === "number")
        return val;
    const { strict } = mergeOptions(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string")
        return undefined;
    const nr = Number.parseFloat(val);
    if (Number.isNaN(nr))
        return undefined;
    return nr;
}, {
    min: (val, o) => val >= o,
    max: (val, o) => val <= o,
    onlyInteger: (val, o) => !o || val % 1 === 0,
    allowUnsafe: (val, o) => o ||
        (val > Number.MIN_SAFE_INTEGER && val < -Number.EPSILON) ||
        (val < Number.MAX_SAFE_INTEGER && val > Number.EPSILON) ||
        val === 0,
    allowNaN: (val, o) => o || !Number.isNaN(val),
}, {
    onlyInteger: false,
    allowUnsafe: false,
    allowNaN: false,
});
const y = new Set().add("1").add(1).add("true");
const n = new Set().add("0").add(0).add("false");
vality.boolean = scalar("boolean", (val, options, context) => {
    if (typeof val === "boolean")
        return val;
    const { strict } = mergeOptions(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string" && typeof val !== "number")
        return undefined;
    return y.has(val) ? true : n.has(val) ? false : undefined;
});
vality.date = scalar("date", (val, options, context) => {
    if (val instanceof Date)
        return val;
    const { strict } = mergeOptions(options, context, ["strict"]);
    if (strict)
        return undefined;
    if (typeof val !== "string" && typeof val !== "number")
        return undefined;
    const date = new Date(val);
    if (Number.isNaN(date.getTime()))
        return undefined;
    return date;
}, {
    min: (val, o) => val >= o,
    max: (val, o) => val <= o,
    past: (val, o) => !o || val < new Date(),
    future: (val, o) => !o || val >= new Date(),
}, {
    past: false,
    future: false,
});
vality.any = scalar("any", (val) => val);
vality.never = scalar("never", () => undefined);
vality.literal = (lit) => {
    const scalarFn = (val, options, context) => {
        if (options.default === true) {
            if (val === undefined)
                return lit;
        }
        else {
            delete options.default;
        }
        if (val === lit)
            return lit;
        const { strict } = mergeOptions(options, context, ["strict"]);
        if (strict)
            return undefined;
        if (typeof lit === "string" && typeof val === "number")
            return val.toString() === lit ? lit : undefined;
        if (typeof lit === "number" && typeof val === "string")
            return parseFloat(val) === lit ? lit : undefined;
        if (typeof lit === "boolean" &&
            (typeof val === "string" || typeof val === "number")) {
            if (lit === true && y.has(val))
                return lit;
            if (lit === false && n.has(val))
                return lit;
        }
        return undefined;
    };
    scalarFn[_type] = lit;
    return scalar("literal", scalarFn);
};
//# sourceMappingURL=scalars.js.map